#include "hotp.h"
#include <openssl/hmac.h>
#include <string.h>

/*
 *  Converts a NULL-terminated base32 string into binary data. Maximum `maxlen`
 *  bytes are written. On illegal input or overflow, zero is returned;
 *  otherwise the number of bytes written is returned.
 */
size_t hotp_base32decode(const char *b32, char *binout, size_t maxlen) {
  size_t pos = 0;
  unsigned char c;
  int buf = 0;
  int cnt = 0;
  for (; *b32; b32++) {
    if (*b32 >= 'A' && *b32 <= 'Z') c = *b32 - 'A';
    else if (*b32 >= 'a' && *b32 <= 'z') c = *b32 - 'a';
    else if (*b32 >= '2' && *b32 <= '7') c = *b32 - '2' + 26;
    else if (*b32 == '0') c = 'O' - 'A';
    else if (*b32 == '1') c = 'I' - 'A';
    else if (*b32 == '=') break;
    else break;
    buf <<= 5;
    buf |= c;
    if (cnt >= 3) {
      if (pos == maxlen) return 0;
      cnt -= 3;
      binout[pos++] = buf >> cnt;
      buf &= (1 << cnt) - 1;
    } else {
      cnt += 5;
    }
  }
  if (cnt >= 5 || buf) return 0;
  for (; *b32; b32++) {
    if (*b32 != '=') return 0;
  }
  return pos;
}

/*
 *  Calculates a HMAC-based one-time password and writes it as NULL-terminated
 *  string to the `code` buffer. The buffer needs to contain at least
 *  `digits`+1 bytes memory. Set `evp_md` to EVP_sha1(), EVP_sha256(), or
 *  EVP_sha512(), as defined in <openssl/evp.h> included by <openssl/hmac.h>.
 *  The secret `key` has to be passed in binary form. Returns 0 on success;
 *  otherwise a non-zero value.
 */
int hotp_generate(
  const EVP_MD *evp_md,
  int digits,
  char *key, int keylen,
  uint64_t counter,
  char *code
) {
  int i;
  unsigned char message[8];
  unsigned char digest[EVP_MAX_MD_SIZE];
  unsigned int digest_len;
  int offset;
  int modulus;
  char format[5];  // "%0_i\0"
  uint32_t codenum;
  if (digits < 6 || digits > 8) return -1;
  for (i=7; i>=0; i--) {
    message[i] = counter;
    counter >>= 8;
  }
  if (!HMAC(evp_md, key, keylen, message, 8, digest, &digest_len)) return -1;
  if (digest_len < 20) return -1;
  offset = digest[digest_len-1] & 0x0f;
  modulus = 1;
  for (i=0; i<digits; i++) modulus *= 10;
  codenum = (
    (digest[offset] & 0x7f) << 24 |
    digest[offset+1] << 16 |
    digest[offset+2] << 8 |
    digest[offset+3]
  ) % modulus;
  strcpy(format, "%0_i");
  format[2] = '0' + digits;
  if (snprintf(code, digits+1, format, codenum) < 0) return -1;
  return 0;
}
