/*
 *  asprintf() needs _GNU_SOURCE on Linux
 */
#ifdef __linux__
#define _GNU_SOURCE
#endif

/*
 *  Include directives
 */
#include <security/pam_appl.h>
#include <security/pam_modules.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <errno.h>
#include <syslog.h>
#include "totp.h"

/*
 *  If atomic opening and locking is not available, define
 *  DOTP_NONATOMIC_LOCKING and emit a warning.
 */
#ifndef O_EXLOCK
#warning O_EXLOCK not defined, using DOTP_NONATOMIC_LOCKING
#define DOTP_NONATOMIC_LOCKING
#endif

/*
 *  Limits of keys and OTP codes (key size in bytes, OTP length in digits)
 */
#define DOTP_MAXKEYLEN 64
#define DOTP_MINCODELEN 6
#define DOTP_MAXCODELEN 8

/*
 *  Allowed counter mismatches and penalty time
 */
#define DOTP_GRACE_HOTP 50
#define DOTP_GRACE_TOTP 30
#define DOTP_WINDOW_HOTP (2+DOTP_GRACE_HOTP)
#define DOTP_WINDOW_TOTP (2*(1+DOTP_GRACE_TOTP))
#define DOTP_PENALTYMS 500

/*
 *  Constants for snprintf() to create pathnames for certain files based on
 *  configuration directory and user name
 */
#define DOTP_PATHFORMAT_KEY "%s/%s.key"
#define DOTP_PATHFORMAT_SUCCEEDED "%s/%s.succeeded"
#define DOTP_PATHFORMAT_SEQ "%s/%s.seq"
#define DOTP_PATHFORMAT_SEQTMP "%s/%s.seq.tmp"


/*
 *  Data structure containing configuration path and per-user configuration.
 */
struct dotp_userconf {
  int debug;                 // non-zero if debug logging is enabled
  int timebased;             // 0=HOTP, non-zero=TOTP
  const EVP_MD *algo;        // used algorithm, e.g. EVP_sha1()
  int codelen;               // length of OTP (6, 7, or 8)
  char key[DOTP_MAXKEYLEN];  // key in binary form
  int keylen;                // length of key in bytes
  FILE *seqfile;             // handle for file containing counter value
  struct {
    char *key;               // name of file holding key and configuration
    char *succeeded;         // name of file indicating previous auth. success
    char *seq;               // name of file containing counter value
    char *seqtmp;            // name of temporary file to write counter value
  } filename;
};

/*
 *  Attempts to use the PAM conversation system to prompt for a string (that is
 *  echoed). Returns PAM_SUCCESS and sets `*resp` to a string containing the
 *  response (which needs to be free'd) on success. Otherwise returns a
 *  PAM error value (as returned by the conversation function) and does not
 *  allocate memory.
 */
static int dotp_query(
  char **resp,
  struct pam_conv *conv,
  char *prompt
) {
  struct pam_message msg[1] = {{
    .msg_style = PAM_PROMPT_ECHO_ON,
    .msg = prompt
  }};
  struct pam_message *msgptr[1] = { &msg[0] };
  struct pam_response *pamresp;
  int retval;
  retval = conv->conv(
    1,
    (const struct pam_message **)msgptr,
    &pamresp,
    conv->appdata_ptr
  );
  if (retval != PAM_SUCCESS) {
    *resp = NULL;
    return retval;
  } else {
    *resp = pamresp[0].resp;
    free(pamresp);
    return PAM_SUCCESS;
  }
}

/*
 *  Returns PAM_SUCCESS if the user name is sane; otherwise returns
 *  PAM_USER_UNKNOWN.
 */
static int dotp_sane_username(const char *user) {
  if (!*user) {
    syslog(LOG_ERR, "Unexpected NULL pointer to user name");
    return PAM_USER_UNKNOWN;
  }
  if (!((*user >= 'A' && *user <= 'Z') || (*user >= 'a' && *user <= 'z'))) {
    syslog(LOG_ERR, "User name does not start with A-Z or a-z");
    return PAM_USER_UNKNOWN;
  }
  for (user++; *user; user++) {
    if (!(
      (*user >= '0' && *user <= '9') ||
      (*user >= 'A' && *user <= 'Z') ||
      (*user >= 'a' && *user <= 'z') ||
      (*user == '-') || (*user == '_')
    )) {
      syslog(LOG_ERR, "User name contains an illegal character");
      return PAM_USER_UNKNOWN;
    }
  }
  return PAM_SUCCESS;
}

/*
 *  Attempts to close the file containing the next valid counter value (if
 *  that file is open) and thus releasing any advisory lock held on the file.
 *  Returns PAM_SUCCESS on success or PAM_SYSTEM_ERR on error.
 */
static int dotp_release_seqfile(struct dotp_userconf *userconf) {
  int res;
  if (!userconf->seqfile) return PAM_SUCCESS;
  if (userconf->debug) syslog(LOG_NOTICE,
    "Closing \"%s\"", userconf->filename.seq
  );
  res = fclose(userconf->seqfile);
  userconf->seqfile = NULL;
  if (!res) return PAM_SUCCESS;
  else {
    syslog(LOG_ERR, "Error while closing \"%s\"", userconf->filename.seq);
    return PAM_SYSTEM_ERR;
  }
}

/*
 *  Releases any resources associated with the `userconf` struct. Does not free
 *  the structure `userconf` itself but only associated data. It is safe to
 *  call this function once after dotp_get_userconf() has been called.
 */
static void dotp_finalize_userconf(struct dotp_userconf *userconf) {
  if (userconf->debug) syslog(LOG_NOTICE, "dotp_finalize_userconf() called");
  dotp_release_seqfile(userconf);
  free(userconf->filename.key);
  free(userconf->filename.succeeded);
  free(userconf->filename.seq);
  free(userconf->filename.seqtmp);
}

/*
 *  Returns non-zero if the file with name `userconf->filename.succeeded`
 *  exists, otherwise zero.
 */
static int dotp_succeeded(struct dotp_userconf *userconf) {
  int status;
  struct stat sb;
  status = stat(userconf->filename.succeeded, &sb);
  if (userconf->debug) {
    if (status) syslog(LOG_NOTICE,
      "Calling stat() on \"%s\" was not successful",
      userconf->filename.succeeded
    );
    else syslog(LOG_NOTICE,
      "Calling stat() on \"%s\" was successful",
      userconf->filename.succeeded
    );
  }
  return !status;
}

/*
 *  Given a configuration path and a user name, loads the per-user
 *  configuration into `dotp_userconf`. Returns PAM_SUCCESS on success,
 *  PAM_IGNORE if the user has no key file (but the configuration directory
 *  is readable), or any other PAM error code on failure. In case of
 *  PAM_SUCCESS, it is mandatory to call dotp_finalize_userconf() to free any
 *  allocated memory associated with the per-user configuration.
 */
static int dotp_get_userconf(
  const char *path, const char *user, struct dotp_userconf *userconf, int debug
) {
  int retval;
  char buf[256];  // minimum 256, see code below
  intmax_t n;
  FILE *f = NULL;
  struct stat sb;
  // Zero *userconf such that calling dotp_finalize_userconf(userconf) is safe
  memset(userconf, 0, sizeof(*userconf));
  userconf->debug = debug;
  retval = dotp_sane_username(user);
  if (retval != PAM_SUCCESS) return retval;
  if (asprintf(
    &userconf->filename.key, DOTP_PATHFORMAT_KEY, path, user
  ) < 0) goto dotp_get_userconf_system_err;
  if (asprintf(
    &userconf->filename.succeeded, DOTP_PATHFORMAT_SUCCEEDED, path, user
  ) < 0) goto dotp_get_userconf_system_err;
  if (asprintf(
    &userconf->filename.seq, DOTP_PATHFORMAT_SEQ, path, user
  ) < 0) goto dotp_get_userconf_system_err;
  if (asprintf(
    &userconf->filename.seqtmp, DOTP_PATHFORMAT_SEQTMP, path, user
  ) < 0) goto dotp_get_userconf_system_err;
  if (!userconf->filename.seqtmp) goto dotp_get_userconf_system_err;
  if (userconf->debug) syslog(LOG_NOTICE,
    "Using \"%s\" as key file", userconf->filename.key
  );
  f = fopen(userconf->filename.key, "r");
  if (!f) {
    if (stat(userconf->filename.key, &sb) == 0 || errno != ENOENT) {
      if (userconf->debug) syslog(LOG_NOTICE,
        "Unexpected error while opening \"%s\"", userconf->filename.key
      );
      goto dotp_get_userconf_system_err;
    } else {
      if (userconf->debug) syslog(LOG_NOTICE,
        "File \"%s\" does not exist", userconf->filename.key
      );
      retval = PAM_IGNORE;
      goto dotp_get_userconf_variable_err;
    }
  }
  if (flock(fileno(f), LOCK_SH)) {
    syslog(LOG_ERR, "Could not share-lock \"%s\"", userconf->filename.key);
    goto dotp_get_userconf_system_err;
  }
  if (fscanf(f, "%255s", buf) != 1 || strlen(buf) > 254) {
    syslog(LOG_WARNING,
      "Could not read first entry (\"hotp\" or \"totp\") in \"%s\"", userconf->filename.key
    );
    goto dotp_get_userconf_system_err;
  }
  if (!strcmp(buf, "hotp")) {
    if (userconf->debug) syslog(LOG_NOTICE, "HOTP selected");
    userconf->timebased = 0;
  } else if (!strcmp(buf, "totp")) {
    if (userconf->debug) syslog(LOG_NOTICE, "TOTP selected");
    userconf->timebased = 1;
  } else {
    syslog(LOG_WARNING,
      "First entry in \"%s\" is neither \"hotp\" nor \"totp\" but \"%s\"",
      userconf->filename.key,
      buf
    );
    goto dotp_get_userconf_system_err;
  }
  if (fscanf(f, "%255s", buf) != 1 || strlen(buf) > 254) {
    syslog(LOG_WARNING,
      "Could not read second entry (\"sha1\", \"sha256\", or \"sha512\") in \"%s\"",
      userconf->filename.key
    );
    goto dotp_get_userconf_system_err;
  }
  if (!strcmp(buf, "sha1")) {
    if (userconf->debug) syslog(LOG_NOTICE, "SHA1 selected");
    userconf->algo = EVP_sha1();
  } else if (!strcmp(buf, "sha256")) {
    if (userconf->debug) syslog(LOG_NOTICE, "SHA256 selected");
    userconf->algo = EVP_sha256();
  } else if (!strcmp(buf, "sha512")) {
    if (userconf->debug) syslog(LOG_NOTICE, "SHA512 selected");
    userconf->algo = EVP_sha512();
  } else {
    syslog(LOG_WARNING,
      "Second entry in \"%s\" is none of \"sha1\", \"sha256\", or \"sha512\", but \"%s\"",
      userconf->filename.key,
      buf
    );
    goto dotp_get_userconf_system_err;
  }
  if (fscanf(f, "%ji", &n) != 1 || n<DOTP_MINCODELEN || n>DOTP_MAXCODELEN) {
    syslog(LOG_WARNING,
      "Third entry in \"%s\" needs to be set to 6, 7, or 8",
      userconf->filename.key
    );
    goto dotp_get_userconf_system_err;
  }
  userconf->codelen = n;
  if (fscanf(f, "%255s", buf) != 1 || strlen(buf) > 254) {
    syslog(LOG_WARNING,
      "Could not read fourth entry (base32 encoded key) in \"%s\"",
      userconf->filename.key
    );
    goto dotp_get_userconf_system_err;
  }
  fclose(f);
  f = NULL;  // needed for cleanup below in case of error
  userconf->keylen = hotp_base32decode(buf, userconf->key, DOTP_MAXKEYLEN);
  if (!userconf->keylen) {
    syslog(LOG_WARNING,
      "Base32 decoding of fourth entry (key) in \"%s\" failed",
      userconf->filename.key
    );
    goto dotp_get_userconf_system_err;
  }
  return PAM_SUCCESS;
  dotp_get_userconf_system_err:
  retval = PAM_SYSTEM_ERR;
  dotp_get_userconf_variable_err:
  if (f) fclose(f);
  dotp_finalize_userconf(userconf);
  return retval;
}

/*
 *  Records a failed login attempt. Returns PAM_SUCCESS or a PAM error code.
 */
static int dotp_escalate(struct dotp_userconf *userconf) {
  if (unlink(userconf->filename.succeeded) && errno != ENOENT) {
    syslog(LOG_ERR, "Could not unlink \"%s\"", userconf->filename.succeeded);
    return PAM_SYSTEM_ERR;
  }
  if (userconf->debug) syslog(LOG_NOTICE,
    "Removed file \"%s\" if existent", userconf->filename.succeeded
  );
  return PAM_SUCCESS;
}

/*
 *  Records a successful login attempt.
 */
static void dotp_deescalate(struct dotp_userconf *userconf) {
  int fd;
  fd = open(userconf->filename.succeeded, O_CREAT|O_NOFOLLOW, (mode_t)0600);
  if (fd == -1) {
    syslog(LOG_WARNING,
      "Could not create \"%s\"",
      userconf->filename.succeeded
    );
  } else {
    if (userconf->debug) syslog(LOG_NOTICE,
      "Created file \"%s\"", userconf->filename.succeeded
    );
    close(fd);
  }
}

/*
 *  Gets the next valid counter value for a HOTP/TOTP code and writes it into
 *  the 64 bit integer pointed to by `seqptr`. Leaves the counter file open and
 *  locked with an exclusive advisory lock. The file can be closed by calling
 *  dotp_release_seqfile() or dotp_finalize_userconf(). Returns PAM_SUCCESS on
 *  success or any other PAM error code.
 */
static int dotp_get_seq(struct dotp_userconf *userconf, uint64_t *seqptr) {
  int retval;
  int fd;
  FILE *f;
  intmax_t n;
  retval = dotp_release_seqfile(userconf);
  if (retval != PAM_SUCCESS) return retval;
  if (userconf->debug) syslog(LOG_NOTICE,
    "Opening \"%s\" with advisory exclusive lock", userconf->filename.seq
  );
#ifdef DOTP_NONATOMIC_LOCKING
  fd = open(userconf->filename.seq, O_RDONLY);
#else
  fd = open(userconf->filename.seq, O_RDONLY|O_EXLOCK);
#endif
  if (!fd) {
    syslog(LOG_ERR, "Could not open \"%s\"", userconf->filename.seq);
    return PAM_SYSTEM_ERR;
  }
#ifdef DOTP_NONATOMIC_LOCKING
  if (flock(fd, LOCK_EX)) {
    syslog(LOG_ERR, "Could not lock \"%s\"", userconf->filename.seq);
    return PAM_SYSTEM_ERR;
  }
#endif
  f = fdopen(fd, "r");
  if (!f) {
    syslog(LOG_ERR, "fdopen() call failed");
    close(fd);
    return PAM_SYSTEM_ERR;
  }
  if (fscanf(f, "%ji", &n) != 1) {
    syslog(LOG_ERR,
      "Could not read counter value from \"%s\"", userconf->filename.seq
    );
    fclose(f);
    return PAM_SYSTEM_ERR;
  }
  userconf->seqfile = f;
  *seqptr = n;
  if (userconf->debug) syslog(LOG_NOTICE, "Counter value read");
  return PAM_SUCCESS;
}

/*
 *  Sets the next valid counter value for the HOTP or TOTP algorithm. Returns
 *  PAM_SUCCESS or a PAM error code.
 */
static int dotp_set_seq(struct dotp_userconf *userconf, uint64_t seq) {
  int fd;
  FILE *f;
  if (userconf->debug) syslog(LOG_NOTICE,
    "Opening \"%s\" to store new counter value", userconf->filename.seqtmp
  );
#ifdef DOTP_NONATOMIC_LOCKING
  fd = open(
    userconf->filename.seqtmp, O_CREAT|O_WRONLY|O_TRUNC, (mode_t)0600
  );
#else
  fd = open(
    userconf->filename.seqtmp, O_CREAT|O_WRONLY|O_TRUNC|O_EXLOCK, (mode_t)0600
  );
#endif
  if (!fd) {
    syslog(LOG_ERR,
      "Could not open \"%s\" for writing", userconf->filename.seqtmp
    );
    return PAM_SYSTEM_ERR;
  }
#ifdef DOTP_NONATOMIC_LOCKING
  if (flock(fd, LOCK_EX)) {
    syslog(LOG_ERR, "Could not lock \"%s\"", userconf->filename.seqtmp);
    return PAM_SYSTEM_ERR;
  }
#endif
  f = fdopen(fd, "w");
  if (!f) {
    syslog(LOG_ERR, "fdopen() call failed");
    close(fd);
    return PAM_SYSTEM_ERR;
  }
  if (fprintf(f, "%ji", (intmax_t)seq) < 0) {
    syslog(LOG_ERR,
      "Could not write to \"%s\"", userconf->filename.seqtmp
    );
    fclose(f);
    return PAM_SYSTEM_ERR;
  }
  if (fflush(f)) {
    syslog(LOG_ERR,
      "Could not write to \"%s\"", userconf->filename.seqtmp
    );
    fclose(f);
    return PAM_SYSTEM_ERR;
  }
  if (rename(userconf->filename.seqtmp, userconf->filename.seq)) {
    syslog(LOG_ERR,
      "Could not rename \"%s\" to \"%s\"",
      userconf->filename.seqtmp,
      userconf->filename.seq
    );
    fclose(f);
    return PAM_SYSTEM_ERR;
  }
  if (fclose(f)) {
    syslog(LOG_ERR,
      "Could not close \"%s\"", userconf->filename.seqtmp
    );
    return PAM_SYSTEM_ERR;
  }
  return PAM_SUCCESS;
}

/*
 *  Checks if a single OTP code is valid, in which case PAM_SUCCESS is
 *  returned. Returns any other PAM error code on failure. Leaves counter file
 *  open and locked, which can be released with dotp_release_seqfile().
 */
static int dotp_check1(struct dotp_userconf *userconf, char *code) {
  char corr[DOTP_MAXCODELEN+1];
  int retval;
  uint64_t valid_seq, seq;
  retval = dotp_get_seq(userconf, &valid_seq);
  if (retval != PAM_SUCCESS) return retval;
  if (userconf->timebased) {
    seq = totp_timeslot(-1);
    if (!seq) {
      syslog(LOG_ERR, "Could not get current time slot");
      return PAM_SYSTEM_ERR;
    }
    if (userconf->debug) syslog(LOG_NOTICE,
      "Using counter value (time slot) %ji for current time", (intmax_t)seq
    );
    if (seq < valid_seq) {
      if (userconf->debug) syslog(LOG_NOTICE,
        "Current counter value has already been used"
      );
      return PAM_AUTH_ERR;
    }
  } else {
    seq = valid_seq;
  }
  if (hotp_generate(
    userconf->algo,
    userconf->codelen,
    userconf->key,
    userconf->keylen,
    seq,
    corr
  )) {
    syslog(LOG_ERR, "Could not calculate correct OTP code");
    return PAM_SYSTEM_ERR;
  }
  if (strcmp(code, corr)) {
    if (userconf->debug) syslog(LOG_NOTICE, "OTP codes do not match");
    return PAM_AUTH_ERR;
  }
  if (userconf->debug) syslog(LOG_NOTICE, "OTP codes match");
  retval = dotp_set_seq(userconf, seq+1);
  if (retval != PAM_SUCCESS) return retval;
  return PAM_SUCCESS;
}

/*
 *  Checks if two consecutive OTP codes (within a greater grace window) are
 *  valid, in which case PAM_SUCCESS is returned. Returns any other PAM error
 *  code on failue. Leaves counter file open and locked, which can be released
 *  with dotp_release_seqfile().
 */
static int dotp_check2(struct dotp_userconf *userconf, char **codes) {
  int retval;
  char corr[2][DOTP_MAXCODELEN+1];
  uint64_t valid_seq, seq;
  int idx, window;
  retval = dotp_get_seq(userconf, &valid_seq);
  if (retval != PAM_SUCCESS) return retval;
  if (userconf->timebased) {
    if (userconf->debug) syslog(LOG_NOTICE,
      "Lowest recorded valid counter value is %ji", (intmax_t)valid_seq
    );
    seq = totp_timeslot(0) - (DOTP_GRACE_TOTP+1);
    if (!seq) {
      syslog(LOG_ERR, "Could not get current time slot");
      return PAM_SYSTEM_ERR;
    }
    if (userconf->debug) syslog(LOG_NOTICE,
      "Lowest allowed time-based counter value for current time is %ji",
      (intmax_t)seq
    );
    window = DOTP_WINDOW_TOTP;
    if (seq < valid_seq) {
      if (valid_seq - seq > DOTP_WINDOW_TOTP-2) {
        if (userconf->debug) syslog(LOG_NOTICE, "Window not big enough");
        return PAM_AUTH_ERR;
      }
      window -= valid_seq - seq;
      seq = valid_seq;
    }
    if (userconf->debug) syslog(LOG_NOTICE,
      "Using lowest counter value %ji with a window size of %i",
      (intmax_t)seq, window
    );
  } else {
    window = DOTP_WINDOW_HOTP;
    seq = valid_seq;
    if (userconf->debug) syslog(LOG_NOTICE, "Using window size of %i", window);
  }
  for (idx=0; idx<window; idx++) {
    if (hotp_generate(
      userconf->algo,
      userconf->codelen,
      userconf->key,
      userconf->keylen,
      seq+idx,
      corr[idx%2]
    )) return PAM_SYSTEM_ERR;
    if (idx) {
      if (
        !strcmp(codes[0], corr[(idx-1)%2]) &&
        !strcmp(codes[1], corr[idx%2])
      ) {
        if (userconf->debug) syslog(LOG_NOTICE,
          "Code pair for counter values %ji and %ji (offset %i and %i) matched",
          (intmax_t)(seq+idx-1), (intmax_t)(seq+idx), idx-1, idx
        );
        retval = dotp_set_seq(userconf, seq+idx+1);
        if (retval != PAM_SUCCESS) return retval;
        return retval;
      }
    }
  }
  if (userconf->debug) syslog(LOG_NOTICE,
    "No OTP code pair in window matched"
  );
  return PAM_AUTH_ERR;
}

/*
 *  Sub-function for pam_sm_authenticate(). Requires a `userconf` pointer to
 *  a per-user configuration, a `conv` pointer to a `struct pam_conv`, and a
 *  pointer to an array holding two string pointers which will be free'd when
 *  set by dotp_authenticate().
 */
int dotp_authenticate(
  struct dotp_userconf *userconf,
  struct pam_conv *conv,
  char **resp
) {
  int retval;
  if (dotp_succeeded(userconf)) {
    if (userconf->debug) syslog(LOG_NOTICE, "Requesting single code first");
    retval = dotp_query(&resp[0], conv, "Code: ");
    if (retval != PAM_SUCCESS) {
      if (userconf->debug) syslog(LOG_NOTICE, "dotp_query() failed");
      return retval;
    }
    if (dotp_succeeded(userconf)) {
      retval = dotp_check1(userconf, resp[0]);
      if (retval == PAM_SUCCESS) {
        if (userconf->debug) syslog(LOG_NOTICE, "Single code succeeded");
        retval = dotp_release_seqfile(userconf);
        return retval;
      } else {
        if (userconf->debug) syslog(LOG_NOTICE, "Single code failed");
      }
      retval = dotp_escalate(userconf);
      if (retval != PAM_SUCCESS) return retval;
      retval = dotp_release_seqfile(userconf);
      if (retval != PAM_SUCCESS) return retval;
      if (userconf->debug) syslog(LOG_NOTICE, "Requesting second code");
    } else {
      if (userconf->debug) syslog(LOG_NOTICE,
        "Concurrent authentication failue"
      );
    }
  } else {
    if (userconf->debug) syslog(LOG_NOTICE,
      "Single code unavailable, requesting two codes"
    );
    retval = dotp_query(&resp[0], conv, "Code 1/2: ");
    if (retval != PAM_SUCCESS) {
      if (userconf->debug) syslog(LOG_NOTICE, "dotp_query() failed");
      return retval;
    }
  }
  retval = dotp_query(&resp[1], conv, "Code 2/2: ");
  if (retval != PAM_SUCCESS) {
    if (userconf->debug) syslog(LOG_NOTICE, "dotp_query() failed");
    return retval;
  }
  retval = dotp_check2(userconf, resp);
  if (userconf->debug) syslog(LOG_NOTICE,
    "Sleeping for %i milliseconds", DOTP_PENALTYMS
  );
  usleep(1000 * DOTP_PENALTYMS);
  if (retval != PAM_SUCCESS) return retval;
  retval = dotp_release_seqfile(userconf);
  if (retval != PAM_SUCCESS) return retval;
  dotp_deescalate(userconf);
  return retval;
}

/*
 *  Function pam_sm_authenticate() for PAM. Expects the first argument
 *  `argv[0]` set to the configuration directory path. Attempts to ask for a
 *  single OTP code first. On failure, requires two consecutive OTP codes until
 *  a successful login was made. Returns PAM_IGNORE if the user has no key file
 *  ("<username>.key") in the configuration directory. If second argument
 *  `argv[1]` is set to "forecefail", always returns PAM_AUTH_ERR if user has a
 *  key file.
 */
int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, const char **argv) {
  int forcefail = 0;
  int debug = 0;
  const char *configpath = NULL;
  int retval;
  const char *user;
  struct dotp_userconf userconf;
  struct pam_conv *conv;
  char *resp[2] = {NULL, NULL};
  {
    int i;
    for (i=0; i<argc; i++) {
      if (!strncmp(argv[i], "dir=", 4)) {
        configpath = argv[i] + 4;
      }
      else if (!strcmp(argv[i], "forcefail")) forcefail = 1;
      else if (!strcmp(argv[i], "debug")) debug = 1;
      else {
        syslog(LOG_ERR, "Error in pam_double_otp: Unexpected option: %s", argv[i]);
        return PAM_SYSTEM_ERR;
      }
    }
  }
  if (debug) syslog(LOG_NOTICE, "pam_double_otp debugging enabled");
  retval = pam_get_user(pamh, &user, NULL);
  if (debug) syslog(LOG_NOTICE,
    "pam_get_user() returned with: %s", pam_strerror(pamh, retval)
  );
  if (retval != PAM_SUCCESS) return retval;
  if (debug) syslog(LOG_NOTICE, "Loading userconf for user \"%s\"", user);
  retval = dotp_get_userconf(configpath, user, &userconf, debug);
  if (debug) syslog(LOG_NOTICE,
    "dotp_get_userconf() returned with: %s", pam_strerror(pamh, retval)
  );
  if (retval != PAM_SUCCESS) return retval;
  if (forcefail) {
    if (debug) syslog(LOG_NOTICE, "fail option selected");
    dotp_finalize_userconf(&userconf);
    return PAM_AUTH_ERR;
  }
  retval = pam_get_item(pamh, PAM_CONV, (const void **) &conv);
  if (debug) syslog(LOG_NOTICE,
    "pam_get_item() for PAM_CONV returned with: %s", pam_strerror(pamh, retval)
  );
  if (retval != PAM_SUCCESS) {
    dotp_finalize_userconf(&userconf);
    return PAM_AUTH_ERR;
  }
  retval = dotp_authenticate(&userconf, conv, resp);
  if (debug) syslog(LOG_NOTICE,
    "dotp_authenticate() returned with: %s", pam_strerror(pamh, retval)
  );
  dotp_finalize_userconf(&userconf);
  free(resp[0]);
  free(resp[1]);
  return retval;
}

/*
 *  Function pam_sm_setcred() for PAM. Does nothing. Returns PAM_SUCCESS.
 */
int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv) {
  return PAM_SUCCESS;
}
